-module(customer_initiator).
-include("customer_initiator.hrl").
-behaviour(gen_fsm).

%% API

propose(Initiator, Offer) ->            
    gen_fsm:send_event(Initiator, {propose, self(), whereis(truck), Offer}).

refuse(Initiator, Reason) ->
    gen_fsm:send_event(Initiator, {refuse, self(), whereis(truck), Reason}).

update(Initiator, Pos, Val) ->
    gen_fsm:send_event(Initiator, {update, Pos, Val}).


%% Callbacks
%% 

start(Contract) ->
    %% no registration
    gen_fsm:start(?MODULE, Contract, []).

reset(Contract,AfterTime) ->
    timer:sleep(timer:seconds(AfterTime)),
    reset(Contract).

reset(Contract) ->
    gen_fsm:start(?MODULE, Contract, []),
    {stop, normal, []}.

init(Contract) -> 
    process_flag(trap_exit, true),
    State = prepare_cfps(Contract),
    {ok, handle_responses, State, _Timeout = timer:seconds(10)}.
        

prepare_cfps(Contract) ->
    %% List the trucks
    Trucks = df:list_trucks(),
    case Trucks of
        [] ->
            %% No trucks in the df
            %% Retry in 10 secs
            timer:sleep(timer:seconds(10)),
            prepare_cfps(Contract);
        _ ->
            SendCfp = fun(T) -> gen_server:cast(T, {cfp, self(), Contract}) end,

            %% For each truck send cfp
            %% Cast the cfp to the truck server
            lists:foreach(SendCfp, Trucks),                                     
            

            %% add the trucks to the state
            State = #state{trucks_left = Trucks,
                           proposers = [],
                           best_proposer = undefined,
                           best_offer = infinity,
                           refuses_cap = 0,
                           refuses_base = 0,
                           contract = Contract},
            State
    end.




%% States
%% 

%% Handle Response Timeout
handle_responses(timeout, State = #state{proposers = Proposers, best_proposer = BestProposer, contract = Contract}) ->
    case BestProposer of
        undefined ->
            %% FULL TIMEOUT
            %% no offer
            reset(Contract);
        _ ->
            %% we have a best offer
            %% send the accept_proposal and the reject_proposals
            send_acceptances(BestProposer, Proposers),
            {next_state, handle_inform, State}
    end;


%% Handle Refuse
handle_responses({refuse, _Refuser, Truck, Reason}, State = #state{trucks_left = TrucksLeft, 
                                                                   proposers = Proposers, 
                                                                   best_proposer = BestProposer, 
                                                                   refuses_cap = RefusesCap,
                                                                   refuses_base = RefusesBase,
                                                                   contract = Contract}) ->

    NewTrucksLeft = lists:delete(Truck,TrucksLeft),

    NewRefusesCap = case Reason == exceeds_capacity of
                        true ->
                            RefusesCap + 1;
                        false ->
                            RefusesCap
                    end,
    NewRefusesBase = case Reason == returning_to_base of
                         true ->
                             RefusesBase + 1;
                         false ->
                             RefusesBase
                    end,

    case NewTrucksLeft of
        [] ->
            %% finished receiving msgs
            
            %% check for at least one offer
            case BestProposer of
                undefined ->
                    %% no offer
                    %% got FULL REFUSES
                    case NewRefusesCap > NewRefusesBase of
                        true ->
                            %% Break the CFP in 2 pieces
                            break_cfp(Contract);
                        false ->
                            %% Reset after some time
                            reset(Contract, 10) 
                    end;
                _ ->
                    %% we have a best offer
                    %% send the accept_proposal and the reject_proposals
                    send_acceptances(BestProposer, Proposers),
                    {next_state, handle_inform, State}
            end;
        _ ->
            %% Trucks Left
            %% Loop for the rest trucks
            NewState = State#state{trucks_left = NewTrucksLeft, refuses_cap = NewRefusesCap, refuses_base = NewRefusesBase},
            {next_state, handle_responses, NewState, _Timeout = timer:seconds(10)}
    end;

%% Handle Propose
handle_responses({propose, Proposer, Truck, Offer}, State = #state{trucks_left = TrucksLeft, 
                                                                    proposers = Proposers, 
                                                                    best_proposer = BestProposer, 
                                                                    best_offer = BestOffer}) ->

    NewTrucksLeft = lists:delete(Truck,TrucksLeft),
    NewProposers = [Proposer | Proposers],
    
    %% RANKING the new proposer
    case Offer < BestOffer of
        true ->
            %% The new proposer won
            NewBestProposer = Proposer,
            NewBestOffer = Offer;
        false ->
            %% The proposer lost
            %% No change
            NewBestProposer = BestProposer,
            NewBestOffer = BestOffer
    end,

    case NewTrucksLeft of
        [] ->
            %% finished receiving msgs
            %% we have a best offer
            %% send the accept_proposal and the reject_proposals
            send_acceptances(NewBestProposer, NewProposers),
            {next_state, handle_inform, State};

        _ ->
            %% Trucks Left
            %% Loop for the rest trucks
            NewState = #state{trucks_left = NewTrucksLeft,
                              proposers = NewProposers,
                              best_proposer = NewBestProposer,
                              best_offer = NewBestOffer},
            {next_state, handle_responses, NewState, _Timeout = timer:seconds(10)}
    end.


%% Handle inform
handle_inform({inform, _TimeElapsed}, State = #state{contract = {_P, V}}) ->
    customer:remove_value(V),
    {stop, normal, State};

%% Handle failure
handle_inform({failure, _Reason}, _State = #state{contract = Contract}) ->
    %% reset the cfp initiator
    reset(Contract);

%% Updates
handle_inform({update, Pos, Val}, State) ->
    NewState = State#state{truck_position = Pos, truck_value = Val},
    {next_state, handle_inform, NewState};

%% Handle Out Of Sequence
handle_inform(_Other, State) ->
    %% disgard them
    {next_state, handle_inform, State}.


%% Other msgs
%% 

%% Truck Responder down
handle_info({'EXIT', _Pid, _Reason}, _StateName, _State = #state{contract = Contract} ) -> % noconnection, shutdown
    reset(Contract).

%% Cleanup
%% 
terminate(normal, _StateName, _StateData) ->    
    ok.
    
%% Extra functions
%% 

send_acceptances(BestProposer, Proposers) ->
    gen_fsm:send_event(BestProposer, accept_proposal),
    link(BestProposer),
    lists:foreach(fun (T) -> gen_fsm:send_event(T, reject_proposal) end,
                  lists:delete(BestProposer,Proposers)).

break_cfp(_Contract = {P,V}) ->
    case V > 1 of
        true ->
            start({P, V/2}),                            % start 2 new inititiators
            start({P, V/2});
        false ->
            start({P, V})                       % small quantity, don't break, just reset
    end,
    {stop, normal, []}.

