-module(truck).
-include("truck.hrl").
-behaviour(gen_server).

%% Accessors
get_value() ->
    gen_server:call(?MODULE, get_value).

get_position() ->
    gen_server:call(?MODULE, get_position).

set_value(V) ->
    gen_server:cast(?MODULE, {set_value, V}).

set_position(P) ->
    gen_server:cast(?MODULE, {set_position, P}).


%% API
%% 

start_link() ->
    {ok, Lat} = application:get_env(?MODULE,lat), 
    {ok, Lng} = application:get_env(?MODULE,lng), 
    {ok, Val} = application:get_env(?MODULE,val),
    {ok, Cap} = application:get_env(?MODULE,cap),
    start_link({Lat,Lng}, Val, Cap).

start_link(Pos, Val, Cap) ->
    State = #state{position = Pos, value = Val, cap = Cap,  status = ?IDLE, jobs = []},
    gen_server:start_link({local, ?MODULE}, ?MODULE, State, []).

stop() ->
    gen_server:cast(?MODULE, stop).

login() ->
    %% the login name is the node
    gen_server:call(?MODULE, login).

logout() ->
    gen_server:cast(?MODULE, logout).

get_future_state() ->
    gen_server:call(?MODULE, get_future_state).

add_goal(From, Contract, FTime, JobTime) ->
    gen_server:call(?MODULE, {add_goal, From, Contract, FTime, JobTime}).

step_done(Step) ->
    gen_server:cast(?MODULE, {step_done, Step}).

job_done(TimeElapsed) ->
    gen_server:cast(?MODULE, {job_done, TimeElapsed}).


%% Initialization
%% 

init(State) ->
    {ok, State}.


%% Calls
%% 

handle_call(get_value, _From, State = #state{value=V}) ->
    {reply, V, State};
handle_call(get_position, _From, State = #state{position=P}) ->
    {reply, P, State};
handle_call(login, _From, State = #state{position=P, value=V}) ->
    Reply = df:subscribe(?MODULE,P,V),
    {reply, Reply, State};
handle_call(get_future_state, _From, State) ->
    %% Reply = {EndPosition, EndValue, AfterWholeTime}
    Reply = compute_future_state(State),
    {reply, Reply, State};

handle_call({add_goal, From, _Contract = {Position, Value}, FTime, JobTime}, _Responder, 
            State = #state{jobs = Jobs, status = Status, cap = MaxCapacity}) ->
    case compute_future_state(State) of
        {FPosition, FCapacity,FTime} ->
            %% create the job
            Job = #job{responder = From, start_position = FPosition,  end_position = Position, value = Value, eta = JobTime},


            %% If low capacity ADD THE BASEJOB
            case (FCapacity - Value) / MaxCapacity < 0.10 of
                false ->
                    NewStatus = ?DRIVING,
                    %% Append just the new job
                    NewJobs = lists:append(Jobs,[Job]);
                true ->
                    NewStatus = ?RETURNING,
                    %% low capacity
                    %% add also the return-to-base job
                    BaseValue = MaxCapacity + Value - FCapacity, % will be added
                    BaseEta = gmaps:compute_eta(Position, ?BASE) + gmaps:compute_pickup(BaseValue),
                    BaseJob = #job{start_position = Position, end_position = ?BASE, value = -(BaseValue), eta = BaseEta},
                    NewJobs = lists:append(Jobs,[Job,BaseJob])
            end,


            %% Check to run the job
            case Status of
                ?IDLE ->
                    truck_driver:run_job(Job);
                _ ->
                    ok
            end,


            NewState = State#state{jobs = NewJobs, status = NewStatus},
            {reply, {ok, goal_added}, NewState};
        _ ->
            {reply, {error,race_condition}, State}
    end.





%% Casts
%% 

handle_cast(logout, State) ->
    df:unsubscribe(),
    {noreply, State};

handle_cast({set_value, V}, State = #state{position=P}) ->
    df:update(?MODULE, P, V),
    {noreply, State#state{value=V}};

handle_cast({set_position, P}, State = #state{value=V}) ->
    df:update(?MODULE, P, V),
    {noreply, State#state{position=P}};

handle_cast(Cfp = {cfp, Initiator, _Contract = {_P, V}}, State = #state{cap = MaxCapacity, status = Status}) ->
    if
        MaxCapacity < V ->
            customer_initiator:refuse(Initiator, exceeds_capacity);
        Status == ?RETURNING ->
            customer_initiator:refuse(Initiator, returning_to_base);
        true ->
            %% Spawn a new truck responder
            %% and pass the cfp as it is
            {ok, ResponderPid} = truck_responder:start(),
            truck_responder:pass_cfp(ResponderPid,Cfp)
    end,
    {noreply, State};

handle_cast({step_done, Step}, State = #state{jobs=Jobs}) ->
    {NewPosition, _Duration} = Step,

    %% Talk with the df
    df:update(?MODULE,NewPosition, State#state.value),

    %% Talk with the responder
    [truck_responder:update(Job#job.responder,NewPosition, State#state.value) || Job <- Jobs, Job#job.responder /= undefined],

    NewState = State#state{position = NewPosition},
    {noreply, NewState};

handle_cast({job_done, TimeElapsed}, State = #state{value=Value, jobs=[DoneJob | RestJobs], status = Status}) ->

    %% pickup the quantity from the customer
    %% capacity decreases
    NewValue = Value - DoneJob#job.value,

    %% Talk with the df
    df:update(?MODULE,DoneJob#job.end_position, NewValue),

    case DoneJob#job.responder of
        undefined ->
            %% Base-job done
            ok;
        Responder ->
            %% Remind the responder
            gen_fsm:send_event(Responder, {done, TimeElapsed})
    end,

    %% Check remaining jobs
    case RestJobs of
        [] ->                                   % No jobs left, go to idle
            NewStatus = ?IDLE;
        [NextJob | _] ->                 % Run the next job
            NewStatus = Status,          % Remains the same status
            truck_driver:run_job(NextJob)
    end,

    NewState = State#state{jobs = RestJobs, value = NewValue, status = NewStatus},

    {noreply, NewState};
    
handle_cast(stop, _State) ->
    {stop, normal, _State}.


%% Cleanup
terminate(normal, _) ->
    ok.

%% Extra functions
compute_future_state(#state{value = Value, jobs = [], position = Position}) -> % no jobs
    {Position, Value, 0};                       % so the same position with the same value in 0 time
compute_future_state(#state{value = Value, jobs = Jobs}) ->
    FPosition = (lists:last(Jobs))#job.end_position,
    {FCapacityAcc, FTime} = lists:foldl(fun(Job,{ValueAcc, EtaAcc}) -> 
                                             {ValueAcc+Job#job.value,
                                              EtaAcc+Job#job.eta}  
                                     end, {0,0}, Jobs),
    {FPosition, Value - FCapacityAcc, FTime}.

    
